<?php

namespace Drupal\edw_paragraph_helper\Controller;

use Drupal\Component\Utility\Html;
use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Database\Connection;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Language\LanguageManagerInterface;
use Drupal\Core\Link;
use Drupal\Core\Pager\PagerManagerInterface;
use Drupal\Core\Routing\RouteMatchInterface;
use Drupal\Core\Url;
use Drupal\paragraphs\Entity\Paragraph;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Helper controller.
 */
class ParagraphController extends ControllerBase {

  /**
   * Total errors found.
   *
   * @var int
   */
  protected $error;

  /**
   * The current route match.
   *
   * @var \Drupal\Core\Routing\CurrentRouteMatch
   */
  protected $currentRouteMatch;

  /**
   * The current paragraph object.
   *
   * @var \Drupal\paragraphs\Entity\Paragraph
   */
  protected $paragraph;

  /**
   * The current language manager.
   *
   * @var \Drupal\Core\Language\LanguageManagerInterface
   */
  protected $languageManager;

  /**
   * Check if we can use paragraphs_edit module.
   *
   * @var bool
   */
  protected $paragraphsEditExists;

  /**
   * The pager manager.
   *
   * @var \Drupal\Core\Pager\PagerManagerInterface
   */
  protected $pagerManager;

  /**
   * The database API.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected $database;

  /**
   * {@inheritdoc}
   */
  public function __construct(
    Connection $database,
    EntityTypeManagerInterface $entity_type_manager,
    RouteMatchInterface $current_route_match,
    LanguageManagerInterface $language_manager,
    ModuleHandlerInterface $module_handler,
    PagerManagerInterface $pager_manager,
    $container,
  ) {
    $this->error = 0;
    $this->database = $database;
    $this->entityTypeManager = $entity_type_manager;
    $this->currentRouteMatch = $current_route_match;
    $this->languageManager = $language_manager;
    $this->paragraphsEditExists = $module_handler->moduleExists('paragraphs_edit');
    $this->pagerManager = $pager_manager;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('database'),
      $container->get('entity_type.manager'),
      $container->get('current_route_match'),
      $container->get('language_manager'),
      $container->get('module_handler'),
      $container->get('pager.manager'),
      $container
    );
  }

  /**
   * Paragraphs overview.
   */
  public function types() {
    $build = [];
    $query = $this->database->select('paragraphs_item', 'n')
      ->fields('n', ['type']);
    $query->addExpression('count(type)', 'a_type');
    $query->groupBy('n.type');

    $ppt = $query->execute()->fetchAllKeyed(0, 1);

    $types = $this->entityTypeManager
      ->getStorage('paragraphs_type')
      ->loadMultiple();

    $rows = [];
    foreach ($types as $type) {
      $rows[] = [
        'label' => Link::createFromRoute(
          $type->label(),
          'edw_paragraph_helper.list',
          [
            'paragraphs_type' => $type->id(),
          ]
        ),
        'id' => $type->id(),
        'q' => !empty($ppt[$type->id()]) ? $ppt[$type->id()] : '-',
      ];
    }

    $header = [
      'id' => $this->t('Label'),
      'label' => $this->t('Machine name'),
      '#' => $this->t('#'),
    ];

    $build['caption']['#markup'] = '<h2>Paragraphs types: ' . count($types) . '</h2>';

    $build['table'] = [
      '#type' => 'table',
      '#header' => $header,
      '#rows' => $rows,
    ];
    $build['table']['#attached']['library'][] = 'edw_paragraph_helper/admin';
    return $build;
  }

  /**
   * Overview paragraphs by type.
   */
  public function list() {
    $build = [];

    $paragraphs_type = $this->currentRouteMatch->getRawParameter('paragraphs_type');
    if (!$paragraphs_type) {
      return $build;
    }

    // Number of items per page.
    $limit = 100;

    // Get the current page number.
    $current_page = $this->pagerManager->findPage();

    // Calculate the offset based on the current page.
    $offset = $current_page * $limit;

    $query = $this->database->select('paragraphs_item', 't')
      ->fields('t', ['id'])
      ->condition('t.type', $paragraphs_type);

    $count_query = clone $query;
    $total_count = $count_query->countQuery()->execute()->fetchField();

    $query->range($offset, $limit);
    // Execute the query and get the results.
    $ids = $query->execute()->fetchAllKeyed(0, 0);

    $this->pagerManager->createPager($total_count, $limit);

    $paragraphs = Paragraph::loadMultiple($ids);
    $build[] = [
      '#markup' => '<h2>Paragraph type: `' . $paragraphs_type . '`</h2><h2> Total: ' . $total_count . '</h2>',
    ];
    foreach ($paragraphs as $paragraph) {
      $data = $this->proccessParagraph($paragraph->id());
      $class = $data['errors'] ? ' paragraph-error' : '';
      $build[] = [
        '#type' => 'table',
        '#attributes' => [
          'class' => ['paragraph-results' . $class],
        ],
        '#title' => '[' . count($data['rows']) . ']',
        '#rows' => $data['rows'],
        '#header' => $this->getHeader(),
      ];
    }
    if ($this->error) {
      if ($total_count > $limit) {
        $build[0]['#markup'] .= '<h2>Errors per page: ' . $this->error . '</h2>';
      }
      else {
        $build[0]['#markup'] .= '<h2>Errors: ' . $this->error . '</h2>';
      }
    }
    $build['#attached']['library'][] = 'edw_paragraph_helper/admin';
    $build['pager'][] = ['#type' => 'pager'];
    return $build;
  }

  /**
   * Paragraph detail check.
   */
  public function paragraph() {
    $current_nid = $this->currentRouteMatch->getRawParameter('paragraph');

    $data = $this->proccessParagraph($current_nid);
    $class = $data['errors'] ? ' paragraph-error' : '';

    $build['table'] = [
      '#type' => 'table',
      '#rows' => $data['rows'],
      '#attributes' => [
        'class' => ['paragraph-results' . $class],
      ],
      '#header' => $this->getHeader(),
    ];

    $language = $this->languageManager->getCurrentLanguage()->getId();
    if ($this->paragraph->hasTranslation($language)) {
      $this->paragraph = $this->paragraph->getTranslation($language);
    }

    $rows = [];
    foreach ($this->paragraph->getFields() as $name => $field) {
      if (strpos($name, 'field_') === 0) {
        $name = '<b>' . $name . '</b>';
      }
      $type = $field->getFieldDefinition()->getType();
      $value = Html::escape($field->getString());
      $rows[] = [
        ['data' => ['#markup' => $name]],
        ['data' => ['#markup' => $value]],
        ['data' => ['#markup' => $type]],
      ];
    }

    $build['fields'] = [
      '#type' => 'table',
      '#title' => '[' . count($rows) . ']',
      '#rows' => $rows,
      '#header' => [
        ['data' => ['#markup' => 'Name']],
        ['data' => ['#markup' => 'Value']],
        ['data' => ['#markup' => 'Type']],
      ],
    ];
    $build['#attached']['library'][] = 'edw_paragraph_helper/admin';
    return $build;
  }

  /**
   * Result table header columns.
   */
  protected function getHeader() {
    $header = [];
    $header[] = ['data' => ['#markup' => 'Type']];
    $header[] = ['data' => ['#markup' => 'Bundle']];
    $header[] = ['data' => ['#markup' => 'Field']];
    $header[] = ['data' => ['#markup' => 'Id']];
    $header[] = ['data' => ['#markup' => 'Revision id']];
    $header[] = ['data' => ['#markup' => 'Revision error']];
    $header[] = ['data' => ['#markup' => 'Label']];
    $header[] = ['data' => ['#markup' => 'Status']];
    $header[] = ['data' => ['#markup' => 'Operations']];
    return $header;
  }

  /**
   * Helper function to process paragraph.
   */
  private function proccessParagraph($current_nid) {
    $items = [];
    if (!empty($current_nid)) {
      $object = Paragraph::load($current_nid);
      $this->paragraph = $object;

      $error = FALSE;
      $parent_id = $object->get('parent_id')->getValue()[0]['value'] ?? '';
      $extra = '';
      // @todo custom functionality
      if ($object->hasField('field_view')) {
        $view = $object->get('field_view')->getValue()[0];
        $extra .= $view['target_id'] . ' ' . $view['display_id'];
      }
      $items[] = [
        'id' => $object->id(),
        'revision_id' => $object->getRevisionId(),
        'parent_id' => $parent_id,
        'bundle' => $object->bundle(),
        'type' => $object->getEntityTypeId(),
        'error' => $error,
        'label' => $object->label(),
        'field' => $extra,
        'rev_error' => '',
        'status' => '',
      ];
      $any_errors = FALSE;
      $entity_type = '';
      $node_id = NULL;
      while (TRUE) {
        $id = $object->id();
        $parent_id = '';
        if ($object->hasField('parent_id')) {
          $parent_id = $object->get('parent_id')->getValue()[0]['value'] ?? '';
        }
        $parent_type = '';
        if ($object->hasField('parent_type')) {
          $parent_type = $object->get('parent_type')->getValue()[0]['value'] ?? '';
        }
        $parent_field_name = '';
        if ($object->hasField('parent_field_name')) {
          $parent_field_name = $object->get('parent_field_name')->getValue()[0]['value'] ?? '';
        }
        if (!$parent_id || !$parent_field_name) {
          $any_errors = TRUE;
          $this->error++;
          break;
        }
        $root_parent_type = $parent_type;

        $revision_id = $object->getRevisionId();
        $object = $object->getParentEntity();
        $rev_error = '';
        $field_name = '';
        if (is_object($object)) {
          $fields = array_keys($object->getFields());
          foreach ($fields as $field) {
            // Skip non field.
            if (strpos($field, 'field_') !== 0) {
              continue;
            }
            $values = $object->get($field)->getValue();
            foreach ($values as $value) {
              if (empty($value['target_id'])) {
                continue;
              }
              if ($value['target_id'] == $id) {
                $field_name .= $field . ' ';
                if (!empty($value['target_revision_id']) && $value['target_revision_id'] != $revision_id) {
                  $rev_error = $value['target_revision_id'];
                }
              }
            }
          }
        }

        $error = FALSE;
        if (!$object) {
          $any_errors = TRUE;
          $this->error++;
          break;
        }
        $entity_type = $object->getEntityTypeId();
        $parent_values = [];
        // @todo check for broken relation...
        if ($object->hasField($parent_field_name)) {
          $parent_values = $object->get($parent_field_name)->getValue();
        }

        $parent_values = array_map(function ($value) {
          return (string) $value['target_id'];
        }, $parent_values);
        if (!in_array($id, $parent_values)) {
          $error = TRUE;
          $any_errors = TRUE;
          $this->error++;
        }
        $status = '';
        if ($entity_type == 'node') {
          $status = $object->get('status')->value ? $this->t('Published') : $this->t('Draft');
          $node_id = $object->id();
        }
        $items[] = [
          'id' => $object->id(),
          'revision_id' => $object->getRevisionId(),
          'parent_id' => $parent_id,
          'bundle' => $object->bundle(),
          'type' => $entity_type,
          'error' => $error,
          'label' => $object->label(),
          'field' => $field_name,
          'rev_error' => $rev_error,
          'status' => $status,
        ];
        if ($entity_type != 'paragraph') {
          break;
        }
        if (in_array($entity_type, ['block_content', 'node', 'taxonomy_term'])) {
          break;
        }
      }
    }
    if ($entity_type == 'paragraph' && !$any_errors) {
      $any_errors = TRUE;
      $this->error++;
    }

    $rows = [];
    foreach ($items as $item) {
      $row = [];
      $operations = [];
      switch ($item['type']) {
        case 'node':
          $operations['view'] = [
            'title' => $this->t('View'),
            'url' => Url::fromRoute(
              'entity.node.canonical',
              [
                'node' => $item['id'],
              ]
            ),
          ];
          $operations['edit'] = [
            'title' => $this->t('Edit'),
            'url' => Url::fromRoute(
              'entity.node.edit_form',
              [
                'node' => $item['id'],
              ]
            ),
          ];
          break;

        case 'block_content':
          $operations['edit'] = [
            'title' => $this->t('Edit'),
            'url' => Url::fromRoute(
              'entity.block_content.canonical',
              [
                'block_content' => $item['id'],
              ]
            ),
          ];
          break;

        case 'taxonomy_term':
          $operations['view'] = [
            'title' => $this->t('View'),
            'url' => Url::fromRoute(
              'entity.taxonomy_term.canonical',
              [
                'taxonomy_term' => $item['id'],
              ]
            ),
          ];
          $operations['edit'] = [
            'title' => $this->t('Edit'),
            'url' => Url::fromRoute(
              'entity.taxonomy_term.edit_form',
              [
                'taxonomy_term' => $item['id'],
              ]
            ),
          ];
          break;

        case 'paragraph':
          $operations['paragraph'] = [
            'title' => $this->t('View'),
            'url' => Url::fromRoute(
              'edw_paragraph_helper.paragraph',
              [
                'paragraph' => $item['id'],
              ]
            ),
          ];
          // @todo Check why NULL root_parent.
          if ($this->paragraphsEditExists && $item['parent_id']) {
            $operations['edit'] = [
              'title' => $this->t('Edit'),
              'url' => Url::fromRoute(
                'paragraphs_edit.edit_form',
                [
                  'root_parent' => $node_id,
                  'root_parent_type' => $root_parent_type,
                  'paragraph' => $item['id'],
                ],
              ),
            ];
          }
          break;

        default:
          $message = $this->t('Invalid type: "@type"', ['@type' => '"' . implode('", "', $item['type']) . '"']);
          throw new \InvalidArgumentException($message);
      }

      if ($this->paragraphsEditExists && $item['type'] == 'paragraph' && $item['parent_id']) {
        $operations['delete'] = [
          'title' => $this->t('Delete'),
          'url' => Url::fromRoute(
            'paragraphs_edit.delete_form',
            [
              'root_parent' => $item['parent_id'],
              'root_parent_type' => $root_parent_type,
              'paragraph' => $item['id'],
            ],
          ),
        ];
      }
      $row[] = ['data' => ['#markup' => $item['type']]];
      $row[] = ['data' => ['#markup' => $item['bundle']]];
      $row[] = ['data' => ['#markup' => $item['field']]];
      $row[] = ['data' => ['#markup' => $item['id']]];
      $row[] = ['data' => ['#markup' => $item['revision_id']]];
      $row[] = ['data' => ['#markup' => $item['rev_error']]];
      $row[] = [
        'data' => ['#markup' => $item['label']],
        'class' => ['wrap'],
      ];
      $row[] = ['data' => ['#markup' => $item['status']]];
      $row[] = [
        'data' =>
          [
            '#type' => 'operations',
            '#links' => $operations,
          ],
      ];
      $rows[] = $row;
    }
    return [
      'rows' => $rows,
      'errors' => $any_errors,
    ];
  }

}
